#ifndef _AGNIX_ADI_BLOCK_FLOPPY_H_
#define  _AGNIX_ADI_BLOCK_FLOPPY_H_

#define BASE_REG		0x3F0

#define FLOPPY_SRA		BASE_REG + 0x00
#define FLOPPY_SRB		BASE_REG + 0x01
#define FLOPPY_DOR		BASE_REG + 0x02
#define FLOPPY_TDR		BASE_REG + 0x03
#define FLOPPY_MSR		BASE_REG + 0x04
#define FLOPPY_DSR		BASE_REG + 0x04
#define FLOPPY_FIFO		BASE_REG + 0x05
#define FLOPPY_RSD		BASE_REG + 0x06
#define FLOPPY_DIR		BASE_REG + 0x07
#define FLOPPY_CCR		BASE_REG + 0x07

/*
 * Floppy SRA register bits
 */
#define SRA_DIR			0x01
#define SRA_WP			0x02
#define SRA_INDX		0x04
#define SRA_HDSEL		0x08
#define SRA_TRK0		0x10
#define SRA_STEP		0x20
#define SRA_DRV2		0x40
#define SRA_INT			0x80

/*
 * Floppy SRB register bits
 */
#define SRB_MOT_EN0		0x01
#define SRB_MOT_EN1		0x02
#define SRB_WE			0x04
#define SRB_RDDATA		0x08
#define SRB_WRDATA		0x10
#define SRB_DRIVE_SEL0		0x20

/*
 * Floppy DOR register bits
 */
#define DOR_DRIVE_SEL0		0x01
#define DOR_DRIVE_SEL1		0x02
#define DOR_RESET_NEG		0x04
#define DOR_DMA_GATE_NEG	0x08
#define DOR_MOT_EN0		0x10
#define DOR_MOT_EN1		0x20
#define DOR_MOT_EN2		0x40
#define DOR_MOT_EN3		0x80

/*
 * Floppy TDR register bits
 */
#define TDR_SEL0		0x01
#define TDR_SEL1		0x02

/*
 * Floppy DSR register bits
 */
#define DSR_DRATE_SEL0		0x01
#define DSR_DRATE_SEL1		0x02
#define DSR_PRE_COMP0		0x04
#define DSR_PRE_COMP1		0x08
#define DSR_PRE_COMP2		0x10
#define DSR_POWER_DOWN		0x40
#define DSR_SW_RESET		0x80

/*
 * Floppy MSR register bits
 */
#define MSR_DRV0_BUSY		0x01
#define MSR_DRV1_BUSY		0x02
#define MSR_DRV2_BUSY		0x04
#define MSR_DRV3_BUSY		0x08
#define MSR_CMD_BUSY		0x10
#define MSR_NON_DMA		0x20
#define MSR_DIO			0x40
#define MSR_RQM			0x80

/*
 * Floppy DIR register bits
 */
#define DIR_HIGH_DENS		0x01
#define DIR_DRATE_SEL0		0x02
#define DIR_DRATE_SEL1		0x04
#define DIR_DSK_CHG		0x80

/*
 * Floppy CCR register bits
 */
#define CCR_DRATE_SEL0		0x01
#define CCR_DRATE_SEL1		0x02

#define FLOPPY_REQ_MFM		0x40
#define FLOPPY_REQ_MT		0x80
#define FLOPPY_REQ_SK		0x20

/*
 * Floppy commands
 */
#define FLOPPY_CMD_READ_DATA	0x06
#define FLOPPY_CMD_DUMPREGS	0x0E
#define FLOPPY_CMD_VERSION	0x10
#define FLOPPY_CMD_PARTID	0x18

#define FLOPPY_MAX_POLL_REQ_REPLIES	0x4000
#define FLOPPY_MAX_POLL_RES_REPLIES	0x4000

/* 
 * Floppy versions
 */

#define FLOPPY_VER_8272A	0x072a
#define FLOPPY_VER_82077	0x0077
#define FLOPPY_VER_82078	0x0078
#define FLOPPY_VER_S82078B	0x078b
#define FLOPPY_VER_87306	0x0306
#define FLOPPY_VER_UNKNOWN	0x0000

#define FLOPPY_DMA		2
#define FLOPPY_IRQ		6
#define FLOPPY_BLOCK_SIZE	0x200

int floppy_reset(void);
int floppy_set_defaults(void);
int floppy_request_command(u8 command);
int floppy_result_command(u8 *result_buf, int max_len);
int floppy_interrupt_check(void);
void floppy_interrupt(u32 irq);
void floppy_motor_on(void);
void floppy_motor_off(void);
int floppy_setup_DMA(struct adi_blkreq_s *blkreq);
int floppy_request_read(struct adi_blkreq_s *blkreq);
int floppy_get_version(void);
int floppy_init(void);
int adi_blkdrv_floppy_init(void);

#endif 
