#ifndef _AGNIX_DRIVERS_NETDRV_H_
#define _AGNIX_DRIVERS_NETDRV_H_

#include <agnix/spinlock.h>
#include <agnix/list.h>
#include <agnix/net/net.h>
#include <agnix/adi/net/adi_net_rtl_8029.h>

#define NETDEV_NAME_MAX_LEN	32
#define NETDEV_MAC_LEN		6

struct adi_netdev_s;

struct adi_netdev_ops_s {
    int (*driver_open)(struct adi_netdev_s *netdev);
    int (*driver_close)(struct adi_netdev_s *netdev);
    int (*transmit_frame)(struct frb_s *frb);
    int (*set_recv_mode)(struct adi_netdev_s *netdev, int recv_mode);
    int (*set_mac_addr)(struct adi_netdev_s *netdev, u8 *mac_addr);
    int (*set_broadcast_addr)(struct adi_netdev_s *netdev, u8 *broadcast_addr);
    int (*set_mtu)(struct adi_netdev_s *netdev, int mtu);
    int (*header_build)(struct frb_s *frb, u8 *mac_dest, u8 *mac_source, u16 proto);
    int (*header_parse)(struct frb_s *frb);
};

struct adi_netdev_s {
    int  if_id;
    char if_name[NETDEV_NAME_MAX_LEN];

    struct net_layer_2_netdev_s if_layer_2;
    struct net_layer_3_netdev_s if_layer_3;

    struct adi_netdev_ops_s *if_ops;
    void *if_internal_param;

    struct list_head list;    
    spinlock_t if_lock;
};

struct adi_netdev_s *adi_netdev_alloc(void);
void adi_netdev_free(struct adi_netdev_s *netdev);
struct adi_netdev_s *adi_netdev_find_by_name(char *name);
int adi_register_netdev(struct adi_netdev_s *netdev);
int adi_netdev_init(struct adi_netdev_s *netdev);
int adi_netdrv_init(void);
void adi_netdev_print_interfaces(void);

#endif
