#ifndef _AGNIX_NET_LAYER_2_ARP_H_
#define _AGNIX_NET_LAYER_2_ARP_H_

#include <agnix/net/layer_3/layer_3.h>

#define ARP_MAX_ENTRIES	4096
#define ARP_BITMAP_LEN	(((ARP_MAX_ENTRIES - 1) >> 5) + 1)
#define ARP_MAX_TIMER	1000

#define ARPHRD_ETHER 	1		/* Ethernet 10Mbps		*/

#define	ARPOP_REQUEST	1		/* ARP request			*/
#define	ARPOP_REPLY	2		/* ARP reply			*/
#define	ARPOP_RREQUEST	3		/* RARP request			*/
#define	ARPOP_RREPLY	4		/* RARP reply			*/
#define	ARPOP_InREQUEST	8		/* InARP request		*/
#define	ARPOP_InREPLY	9		/* InARP reply			*/
#define	ARPOP_NAK	10		/* (ATM)ARP NAK			*/

#define ARP_EXPIRE	300
#define ARP_TIMEOUT	300

#define ETH_ALEN	6

struct arp_hdr_s
{
	u16	ar_hrd;			/* format of hardware address	*/
	u16	ar_pro;			/* format of protocol address	*/
	u8	ar_hln;			/* length of hardware address	*/
    	u8	ar_pln;			/* length of protocol address	*/
	u16	ar_op;			/* ARP opcode (command)		*/
	u8	ar_sha[ETH_ALEN];	/* sender hardware address	*/
	u8	ar_sip[4];		/* sender IP address		*/
	u8	ar_tha[ETH_ALEN];	/* target hardware address	*/
	u8	ar_tip[4];		/* target IP address		*/
};

struct arp_entry_s {
    int idx;
    struct arp_neigh_s *neigh;
    u32 timer;
    struct list_head list;
};

struct arp_neigh_s {
    u8 mac[ETH_ALEN];
    struct net_addr_s addr;
};

struct arp_request_s {
    struct list_head list;
    struct arp_neigh_s *neigh;
    struct task_s *task;
};

struct arp_stats_s {
    u32 total_recv_arps;
    u32 recv_request_arps;
    u32 recv_reply_arps;
    u32 total_send_arps;
    u32 send_request_arps;
    u32 send_reply_arps;
};

int net_layer_2_arp_init(void);
int arp_resolve(struct adi_netdev_s *netdev, struct arp_neigh_s *neigh);
int arp_print_stats(void);

#endif
