/*
 * kernel_sys/console.c
 *
 * Copyright (c) 2003-2005 Lukasz Dembinski <dembol@nasa.com.pl>
 * All Rights Reserved
 * 
 * Date:	2005/04
 * Author: 	Lukasz Dembinski
 * Info:
 * Contact:	mailto: <dembol@nasa.com.pl>
 *
 */

#include <agnix/agnix.h>
#include <agnix/adi/adi.h>
#include <agnix/net/net.h>
#include <agnix/terminal.h>
#include <agnix/threads.h>
#include <agnix/sprintf.h>
#include <agnix/memory.h>
#include <agnix/machine.h>
#include <agnix/counters.h>
#include <agnix/sched.h>
#include <agnix/console.h>
#include <agnix/strings.h>
#include <agnix/unistd.h>
#include <agnix/agnix_console.h>

#define MOD_NAME	"CONSOLE: "

extern struct adi_netdev_s rtl_8029_netdev;
extern int counter_jiffies_desc;

struct agnix_console_menu_s * agnix_console_print_help_name(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command)
{
    struct agnix_console_command_s *commands = menu->commands;
    int i, j;

    for (i = 0; i < COMMAND_MAX_ENTRIES; i++) {
	if (commands[i].name == NULL)
	    break;
	
	printf("%s", commands[i].name);
	for (j = strlen(commands[i].name); j < 10; j++)
	    printf(" ");
	
	printf(" - %s\n", commands[i].desc);
    }
    
    return NULL;
}

struct agnix_console_menu_s * agnix_console_print_help(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    return agnix_console_print_help_name(menu, command);
}

struct agnix_console_menu_s * agnix_console_next_node(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    return command->submenu;
}

struct agnix_console_menu_s * agnix_console_prev_node(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    return menu->parent;
}

struct agnix_console_menu_s * agnix_console_ip_stats(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    ip_print_stats();    
    return NULL;
}

struct agnix_console_menu_s * agnix_console_ip_debug(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    int debug;
    
    debug = ip_ioctl_get_debug() ^ 1; 
    ip_ioctl_set_debug(debug);
    
    return NULL;
}

struct agnix_console_menu_s * agnix_console_arp_stats(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    arp_print_stats();    
    return NULL;
}

struct agnix_console_menu_s * agnix_console_arp_request(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    struct arp_neigh_s neigh;
    char *req_addr;

    req_addr = buf;

    if (req_addr == NULL) {
	printf("\nYou must specify IP address for ARP request\n");
	printf("Usage: request [IP]\n");
	return NULL;
    }

    if (net_addr(req_addr, &neigh.addr) < 0) {
        printf("\nInvalid format of interface address\n");
        return NULL;
    }

    neigh.addr.address_len = 4;
    memset(neigh.mac, 0, ETH_ALEN);

    printf("\nWho-has %d.%d.%d.%d\n", 
    neigh.addr.address[0], neigh.addr.address[1], neigh.addr.address[2], neigh.addr.address[3]);

    if (arp_resolve(&rtl_8029_netdev, &neigh) < 0) {
	printf("%d.%d.%d.%d unreachable\n", 
	neigh.addr.address[0], neigh.addr.address[1], neigh.addr.address[2], neigh.addr.address[3]);
    } else {
	printf("%d.%d.%d.%d is-at %02x:%02x:%02x:%02x:%02x:%02x\n",
	neigh.addr.address[0], neigh.addr.address[1], neigh.addr.address[2], neigh.addr.address[3],
	neigh.mac[0], neigh.mac[1], neigh.mac[2], neigh.mac[3], neigh.mac[4], neigh.mac[5]);
    }
    
    return NULL;
}

struct agnix_console_menu_s * agnix_console_pci_list(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    pcibus_print_all();    
    return NULL;
}

struct agnix_console_menu_s * agnix_console_print_memory(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    bootmem_print_memory();    
    return NULL;
}

struct agnix_console_menu_s * agnix_console_print_cpuinfo(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    cpu_print_cpuinfo();
    return NULL;
}

struct agnix_console_menu_s * agnix_console_print_jiffies(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    u32 time[2];
    counter_read(counter_jiffies_desc, time);

    printk("Current jiffies: %08x%08x\n", time[1], time[0]);
    return NULL;
}

struct agnix_console_menu_s * agnix_console_halt(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    do_machine_halt();
    return NULL;
}

struct agnix_console_menu_s * agnix_console_powerdown(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    do_machine_powerdown();
    return NULL;
}

struct agnix_console_menu_s * agnix_console_reboot(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    do_machine_reboot();
    return NULL;
}

struct agnix_console_menu_s * agnix_console_ps(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    scheduler_print_tasks();
    return NULL;
}

struct agnix_console_menu_s * agnix_console_net_ifconfig(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    struct net_addr_s addr;
    struct adi_netdev_s *netdev;
    char *if_name;
    char *if_addr;

    if_name = buf;
    if_addr = word_next(buf, COMMAND_MAX_NAME_LEN);

    if ((if_name != NULL) && (if_addr == NULL)) {
    	printf("\nYou must specify IP address for interface\n");
	printf("Usage: ifconfig [ifname addr]\n");
	return NULL;
    }

    if (if_name != NULL) {
	if ((netdev = adi_netdev_find_by_name(if_name)) == NULL) {
	    printf("\n%s interface not found\n", if_name);
	    return NULL;
	} 
    
	if (net_addr(if_addr, &addr) < 0) {
	    printf("\nInvalid format of interface address\n");
	    return NULL;
        }

        printf("\nSetting %s IP to %d.%d.%d.%d\n", netdev->if_name,
	addr.address[0], addr.address[1], addr.address[2], addr.address[3]);
	addr.address_len = 4;
	
	ip_ioctl_set_ifaddr(netdev, &addr);
    } else {
	adi_netdev_print_interfaces();
    }
     
    return NULL;
}

struct agnix_console_menu_s * agnix_console_vm_stats(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    buddy_alloc_print_stats();
    return NULL;
}

struct agnix_console_menu_s * agnix_console_vm_free(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    buddy_alloc_print_free();
    return NULL;
}

struct agnix_console_menu_s * agnix_console_vm_alloc(struct agnix_console_menu_s *menu, struct agnix_console_command_s *command, char *buf)
{
    u32 *ptr;

    ptr = (u32 *)0xdfffd000;
    printf("%x\n", *ptr);
    return NULL;
}

struct agnix_console_menu_s * agnix_console_command_exec(struct agnix_console_menu_s *menu, char *buf)
{
    struct agnix_console_command_s *command = menu->commands;
    struct agnix_console_menu_s *next_menu;
    char *ptr;
    int i;

    for (i = 0; i < COMMAND_MAX_ENTRIES; i++) {
	if (command[i].name == NULL)
	    break;
	
	if (!word_cmp(command[i].name, buf, COMMAND_MAX_NAME_LEN)) {

	    ptr = word_next(buf, COMMAND_MAX_NAME_LEN);
	    if (command[i].action != NULL) {
	        if ((next_menu = command[i].action(menu, &command[i], ptr)) == NULL)
		    return menu;
	        else
		    return next_menu;
		} else {
		    printf("Command action is NULL\n");
		    return menu;
		}
	}
    }

    printf("Unknown command, try 'help'\n");
    
    return menu;
}

struct agnix_console_menu_s console_main_menu;
struct agnix_console_menu_s console_net_menu;
struct agnix_console_menu_s console_ip_menu;
struct agnix_console_menu_s console_arp_menu;
struct agnix_console_menu_s console_adi_menu;
struct agnix_console_menu_s console_pci_menu;
struct agnix_console_menu_s console_arch_menu;
struct agnix_console_menu_s console_vm_menu;

struct agnix_console_command_s console_ip_commands[] = {
    { "help",   "Print this help",              agnix_console_print_help, &console_ip_menu, NULL },
    { "return", "Return to parent menu",        agnix_console_prev_node,  &console_ip_menu, NULL },
    { "stats",  "Statistics of IP protocol",    agnix_console_ip_stats,   &console_ip_menu, NULL },
    { "debug",  "Enable/disable packets debug", agnix_console_ip_debug,   &console_ip_menu, NULL },
    { NULL }
};

struct agnix_console_command_s console_arp_commands[] = {
    { "help",   "Print this help",            agnix_console_print_help, &console_arp_menu, NULL },
    { "return", "Return to parent menu",      agnix_console_prev_node,  &console_arp_menu, NULL },
    { "request","Send ARP Request",           agnix_console_arp_request,  &console_arp_menu, NULL },
    { "stats",  "Statistics of ARP protocol", agnix_console_arp_stats,  &console_arp_menu, NULL },
    { NULL }
};

struct agnix_console_command_s console_net_commands[] = {
    { "help",   "Print this help",       agnix_console_print_help, &console_net_menu, NULL },
    { "ip",     "IP protocol module",    agnix_console_next_node,  &console_net_menu, &console_ip_menu  },
    { "arp",    "ARP protocol module",   agnix_console_next_node,  &console_net_menu, &console_arp_menu  },
    { "ifconfig", "ifconfig emulator",   agnix_console_net_ifconfig, &console_net_menu, NULL  },
    { "return", "Return to parent menu", agnix_console_prev_node,  &console_net_menu, NULL },
    { NULL }
};

struct agnix_console_command_s console_adi_commands[] = {
    { "help",   "Print this help",       agnix_console_print_help, &console_adi_menu, NULL },
    { "pci",    "PCI Bus module",        agnix_console_next_node,  &console_adi_menu, &console_pci_menu  },
    { "return", "Return to parent menu", agnix_console_prev_node,  &console_adi_menu, NULL },
    { NULL }
};

struct agnix_console_command_s console_pci_commands[] = {
    { "help",   "Print this help",            agnix_console_print_help, &console_pci_menu, NULL },
    { "return", "Return to parent menu",      agnix_console_prev_node,  &console_pci_menu, NULL },
    { "list",   "List all PCI devices",       agnix_console_pci_list,   &console_pci_menu, NULL },
    { NULL }
};

struct agnix_console_command_s console_arch_commands[] = {
    { "cpuinfo", "Print cpu info",        agnix_console_print_cpuinfo, &console_adi_menu, NULL },
    { "help",    "Print this help",       agnix_console_print_help,    &console_adi_menu, NULL },
    { "jiffies", "Print current jiffies", agnix_console_print_jiffies, &console_adi_menu, NULL },
    { "memmap",  "Print memory map",      agnix_console_print_memory,  &console_adi_menu, NULL },
    { "return",  "Return to parent menu", agnix_console_prev_node,     &console_adi_menu, NULL },
    { NULL }
};

struct agnix_console_command_s console_vm_commands[] = {
    { "alloc",    "VM alloc pages",        agnix_console_vm_alloc,   &console_vm_menu, NULL  },
    { "free",     "VM free",               agnix_console_vm_free,    &console_vm_menu, NULL  },
    { "help",     "Print this help",       agnix_console_print_help, &console_vm_menu, NULL },
    { "stats",    "VM statistics",         agnix_console_vm_stats,   &console_vm_menu, NULL  },
    { "return",   "Return to parent menu", agnix_console_prev_node,  &console_vm_menu, NULL },
    { NULL }
};

struct agnix_console_command_s console_main_commands[] = {
    { "adi",       "Driver Interface Subsystem",   agnix_console_next_node,  &console_main_menu,  &console_adi_menu },
    { "arch",      "Architecture dependent subsystem",
					  	   agnix_console_next_node,  &console_main_menu,  &console_arch_menu },
    { "halt",      "Halt machine",                 agnix_console_halt,       &console_main_menu,  NULL },
    { "help",      "Print this help",              agnix_console_print_help, &console_main_menu,  NULL },
    { "net",       "Network subsystem",            agnix_console_next_node,  &console_main_menu,  &console_net_menu },
    { "powerdown", "Powerdown machine",            agnix_console_powerdown,  &console_main_menu,  NULL },
    { "ps",        "Task list",                    agnix_console_ps,         &console_main_menu,  NULL },
    { "reboot",    "Reboot machine",               agnix_console_reboot,     &console_main_menu,  NULL },
    { "vm",        "Virtual memory subsystem",     agnix_console_next_node,  &console_main_menu,  &console_vm_menu },
    { NULL }
};

struct agnix_console_menu_s console_main_menu = { "main", console_main_commands, NULL };
struct agnix_console_menu_s console_adi_menu  = { "adi",  console_adi_commands,  &console_main_menu };
struct agnix_console_menu_s console_pci_menu  = { "pci",  console_pci_commands,  &console_adi_menu };
struct agnix_console_menu_s console_net_menu  = { "net",  console_net_commands,  &console_main_menu };
struct agnix_console_menu_s console_ip_menu   = { "ip",   console_ip_commands,   &console_net_menu };
struct agnix_console_menu_s console_arp_menu  = { "arp",  console_arp_commands,  &console_net_menu };
struct agnix_console_menu_s console_arch_menu = { "arch", console_arch_commands, &console_main_menu };
struct agnix_console_menu_s console_vm_menu   = { "vm",   console_vm_commands,   &console_main_menu };

extern struct terminal_s term_sys[MAX_TERMINALS];
extern int term_cur;

void agnix_console_main(void)
{
    struct agnix_console_menu_s *cur_menu = &console_main_menu;
    char buf[COMMAND_MAX_LEN + 1];
    int ret;

    for (;;) {

        printf("agnix %s> ", cur_menu->name); 

        memset(buf, 0, COMMAND_MAX_LEN + 1);
	if ((ret = read(0, buf, COMMAND_MAX_LEN)) > 0) {
	    cur_menu = agnix_console_command_exec(cur_menu, buf);
	    printf("\n");
	}
    }
}

void agnix_console_thread(void *data)
{
    char console_name[10];
    int console_nr;

    console_nr = (int)data;

    sprintf(console_name, "tty%d", console_nr);
    if ((open(console_name, 0, 0)) < 0) {
	printf("can not open terminal\n");
	for(;;);
    }

    printf(MOD_NAME "Agnix Console %s starting ... ok\n\n", console_name, console_nr);

    agnix_console_main();
}

int agnix_console_start(void)
{
    int i;
    
    for (i = 0; i < MAX_TERMINALS; i++) {    
        create_kernel_thread("console", agnix_console_thread, (void *)i);
    }
    
    return 0;
}
