/*
 * kernel_net/layer_3/ip/ip_core.c
 *
 * Copyright (c) 2003-2005 Lukasz Dembinski <dembol@nasa.com.pl>
 * All Rights Reserved
 * 
 * Date:	2005/06
 * Author: 	Lukasz Dembinski
 * Info:	ip_core.c core file
 * Contact:	mailto: <dembol@nasa.com.pl>
 *
 */

#include <agnix/agnix.h>
#include <agnix/adi/adi.h>
#include <agnix/memory.h>
#include <agnix/net/net.h>
#include <agnix/list.h>
#include <agnix/console.h>

#define MOD_NAME	"NET: "

int ip_recv(struct frb_s *frb);
int ip_debug = 0;

struct net_layer_2_pack ip_pack = {
    .pack_type 		= 0x0008,
    .receiver 		= ip_recv,
};

static struct ip_stats_s ip_stats;

int ip_ioctl_set_ifaddr(struct adi_netdev_s *netdev, struct net_addr_s *addr)
{
    memcpy(&netdev->if_layer_3.if_addr, addr, sizeof(struct net_addr_s));

    return 0;
}

int ip_ioctl_set_debug(int enable)
{
    ip_debug = enable;
    
    return 0;
}

int ip_ioctl_get_debug(void)
{
    return ip_debug;
}

int ip_print_stats(void)
{
    printf("\n");
    printf("Total IP packets:     %d\n", ip_stats.total_packets);
    printf("Bad checksum packets: %d\n", ip_stats.bad_checksum_packets);
    
    return 0;
}

int ip_recv(struct frb_s *frb)
{
    struct ip_hdr_s *iphdr = (struct ip_hdr_s *)frb_data(frb);
    u16 checksum;
    u8 *saddr_ptr = (u8 *)&iphdr->saddr;
    u8 *daddr_ptr = (u8 *)&iphdr->daddr;

    if (ip_debug) {
        printf("Packet IP recv %d.%d.%d.%d -> %d.%d.%d.%d\n", 
	saddr_ptr[0], saddr_ptr[1], saddr_ptr[2], saddr_ptr[3], 
        daddr_ptr[0], daddr_ptr[1], daddr_ptr[2], daddr_ptr[3]);
    }

    ip_stats.total_packets++;

    checksum = ip_checksum(iphdr);

    if (ip_checksum(iphdr) != 0) {
        printk("Packet IP bad checksum\n");
	ip_stats.bad_checksum_packets++;
    }

    return 0;
}

int net_layer_3_ip_init(void)
{
    printk(", IPv4");

    net_layer_2_add_pack(&ip_pack);
    
    return 0;
}
