/*
 * kernel_super/fastirq/fastirq.c
 *
 * Copyright (c) 2003-2004 Lukasz Dembinski <dembol@nasa.com.pl>
 * All Rights Reserved
 * 
 * Date:	2004/10
 * Author: 	Lukasz Dembinski
 * Info:	fastirq.c core file
 * Contact:	mailto: <dembol@nasa.com.pl>
 * 
 */

#include <agnix/agnix.h>
#include <agnix/list.h>
#include <agnix/irq.h>
#include <agnix/queues.h>
#include <agnix/tasks.h>
#include <agnix/threads.h>
#include <agnix/sched.h>
#include <agnix/memory.h>
#include <agnix/ioport.h>
#include <agnix/fastrand.h>
#include <agnix/console.h>

#define MOD_NAME	"FASTIRQS: "

int fastirq_queue_desc;
struct task_s *fastirq_task;
int fastirq_started = 0;
u32 fastirq_count[IRQ_MAX];

extern struct irq_s irq_info[IRQ_MAX];
extern u32 random_val;

int fastirq_queue(u8 irq)
{
    fastirq_count[irq]++;
    sched_activate_task(fastirq_task);

    return 0;
}

void fastirq_run_irq(u8 irq)
{
    struct irq_s *irq_struct = &irq_info[irq];
    struct irq_routine_s *irq_proc;
    struct list_head *tmp;
    void (*proc)(u32);
    int  (*check)(void);
    int  irq_check_ok;
    u32 data;
    u32 flags;

    spin_lock_irqsave(&irq_struct->lock, flags);

    if (!list_empty(&irq_struct->head)) {
	list_for_each(tmp, &irq_struct->head) {
	    irq_proc = list_entry(tmp, struct irq_routine_s, list);
	    proc  = irq_proc->proc;
	    data  = irq_proc->data;
	    check = irq_proc->check; 
	    irq_check_ok = 1;
	    
	    if (check)
	        irq_check_ok = check();
	    
	    if (!irq_check_ok)
		continue;
	    
	    if (proc) {
		proc(data);
	    }

	    if (irq_proc->flags & IRQ_FLAG_RANDOM)
	        random_val = _fastrand(random_val);
	}
    }

    spin_unlock_irqrestore(&irq_struct->lock, flags);
}

void fastirq_thread(void *data)
{
    int i;

    fastirq_task    = current_task;
    fastirq_started = 1;

    __cli();
    
    for (;;) {
	for (i = 0; i < IRQ_MAX; i++) {	
	    if (fastirq_count[i] > 0) {
		fastirq_run_irq((u8)i);
		fastirq_count[i] = 0;
	    }
	}

	sched_deactivate_task(current_task);
	schedule_task();
    }
}

int fastirq_init(void)
{	
    printk(MOD_NAME "Initializing fastirqs\n");

    create_kernel_thread("fastirq", fastirq_thread, NULL);

    return 0;
}
